/* This code may be redistributed and reused; however, all credit
 * must be attributed to the original authors.
 * Authors: Craig Lombardo & Rachel Szteinberg '17
 * Date Created: 12/13/2016
 */

#include "config.h"
#include "tft_master.h"
#include "tft_gfx.h"
#include <stdlib.h>
#include "adc_intf.h"
#include "pt_cornell_1_2.h"
#include<stdio.h>

#include <string.h>

#define	SYS_FREQ 40000000

// string buffers
char buffer[60];
char last_dir_buff[10] = "";
char pass_buff[10] = "";

int i;

//Used for the screens
#define LOCKSCREEN 0
#define MAINSCREEN 1
#define NEWPASSSCREEN 2

int main_option = 0;

int buffPress, buffDir;

//These are the same enumerated codes as can be found in the key code
enum ActionCodes {
    OFF,
    SINGLE,
    DOUBLE,
    HOLD,
    SOUTH,
    NORTH,
    WEST,
    EAST
};
enum ActionCodes actionCode = OFF;

//This is to set basic information for the lock
#define MAXPASSLEN 10
int password[MAXPASSLEN];
int password_attempt[MAXPASSLEN];

//Last direction detected
int lastDir = 0;

//The index of the password attempt
int password_index = 0;

//These are used to help control screen switching
int c_screen = 0;
int drawScreen = 1;
int valid_password = 0;

static struct pt pt_draw_screen;
static struct pt pt_UART_conn, pt_input, pt_output;

// UART command structure variables
char UART_buffer[60];
static struct pt pt_read_UART;

int connection_established = 0;

int buffInput;

/* This method will allow the user to input the last direction seen,
 * and will output a corresponding char.
 * USE THE ENUM CODES OR THIS COULD CAUSE A SEG FAULT
 * Authors: Craig Lombardo & Rachel Szteinberg '17
 */
char* getDirChar(int lastDir) {
    char out;
    switch (lastDir) {
        case NORTH:
            out = 'N';
            break;
        case SOUTH:
            out = 'S';
            break;
        case EAST:
            out = 'E';
            break;
        case WEST:
            out = 'W';
            break;
    }
    return out;
}

//Authors: Brandon Smith and Nick Falco
//delays during init
void delay(unsigned t) { //uses internal timer 2
    T2CON = 0x8000; // enable tmr1, Tcy, 1:1
    while (t--) {
        TMR2 = 0;
        // while (TMR1<8000);
        while (TMR2 < 8000);
    }
}

/* This method will set the color of the LED the same way as it is outlined in
 * the key code.
 * Authors: Craig Lombardo & Rachel Szteinberg '17
 */
static void changeLED(int led) {
    //Red
    if (led >> 2 & 0x1) mPORTASetBits(BIT_3);
    else mPORTAClearBits(BIT_3);

    //Green
    if (led >> 1 & 0x1) mPORTBSetBits(BIT_4);
    else mPORTBClearBits(BIT_4);

    //Blue
    if (led & 0x1) mPORTASetBits(BIT_4);
    else mPORTAClearBits(BIT_4);
}


/* This method will reset the password required to unlock the device.
 * Authors: Craig Lombardo & Rachel Szteinberg '17
 */
void resetPassword() {
    for (i = 0; i < MAXPASSLEN; i++) {
        password[i] = 10;
    }
}

/* This method will reset the password attempt that the user has input thus
 * far.
 * Authors: Craig Lombardo & Rachel Szteinberg '17
 */
void resetPasswordAttempt() {
    for (i = 0; i < MAXPASSLEN; i++) {
        password_attempt[i] = 10;
    }
    sprintf(pass_buff, "");
    password_index = 0;
}

/* This method will compare the required password with the user's input password
 * attempt.
 * Authors: Craig Lombardo & Rachel Szteinberg '17
 */
int checkPassword() {
    int valid = 1;
    for (i = 0; i < MAXPASSLEN && valid; i++) {
        if (password[i] != password_attempt[i]) valid = 0;
    }
    password_index = 0;
    return valid;
}

/* This method will blink the LED to notify the user of an incorrect or correct
 * password attempt. Green for valid, red for invalid.
 * Authors: Craig Lombardo & Rachel Szteinberg '17
 */
void pulseLED(int valid) {
    for (i = 0; i < 4; i++) {
        if (valid) changeLED(2);
        else changeLED(4);
        delay(1000);
        changeLED(1);
        delay(1000);
    }
}

/* This thread will control the screen draw actions.
 * Authors: Craig Lombardo & Rachel Szteinberg '17
 */
static PT_THREAD(protothread_draw_screen(struct pt *pt)) {
    PT_BEGIN(pt);

    while (1) {

        PT_YIELD_TIME_msec(10);

        //Once we get a signal from the bluetooth modules we want to change
        //the color of the LED to signify a valid connection
        if (connection_established) changeLED(5);

        switch (c_screen) {
            case LOCKSCREEN:
                if (drawScreen) {
                    //idle color
                    changeLED(1);

                    tft_fillScreen(ILI9341_BLACK);
                    tft_setCursor(0, 0);
                    drawScreen = 0;
                    tft_setCursor(0, 0);
                    tft_writeString("Hello World\nEnter the PSWRD\n");

                    tft_setCursor(50, 120);
                    tft_writeString("Last Dir:");
                    tft_setCursor(100, 150);
                    tft_writeString(last_dir_buff);

                    //draw_screen = 0 so we only draw this once/prevent flicker
                    drawScreen = 0;
                    main_option = 0;
                    resetPasswordAttempt();
                }
                //If the user holds the button and they have not already entered
                //in the maximum password length
                if ((actionCode > HOLD) && (password_index < MAXPASSLEN)) {
                    //Overwrite the last direction detected
                    tft_setCursor(100, 150);
                    tft_setTextColor(ILI9341_BLACK);
                    tft_writeString(last_dir_buff);

                    //Draw the new direction detected
                    tft_setCursor(100, 150);
                    tft_setTextColor(ILI9341_WHITE);
                    sprintf(last_dir_buff, "%c", getDirChar(actionCode));
                    tft_writeString(last_dir_buff);

                    lastDir = actionCode;
                } 
                //Only enter the user's password if they double click and there
                //is a valid last detection
                else if (actionCode == DOUBLE && lastDir != 0) {
                    tft_setCursor(20, 50);
                    sprintf(pass_buff, "%s%s", pass_buff, "*");
                    tft_writeString(pass_buff);
                    password_attempt[password_index++] = lastDir;


                    tft_setCursor(100, 150);
                    tft_setTextColor(ILI9341_BLACK);
                    tft_writeString(last_dir_buff);

                    tft_setCursor(100, 150);
                    tft_setTextColor(ILI9341_WHITE);
                    
                    //Clear so we cannot 'spam' the input
                    lastDir = 0;
                    sprintf(last_dir_buff, "%s", "None");
                    tft_writeString(last_dir_buff);
                } 
                //Confirmation of a password attempt
                else if (actionCode == SINGLE) {
                    valid_password = checkPassword();
                    pulseLED(valid_password);
                    if (valid_password) {
                        tft_fillScreen(ILI9341_BLACK);
                        tft_setCursor(0, 0);
                        c_screen = MAINSCREEN;
                    } else {
                        tft_fillScreen(ILI9341_BLACK);
                        tft_setCursor(0, 0);
                        drawScreen = 0;
                        tft_setCursor(0, 0);
                        tft_writeString("Hello World\nEnter the PSWRD\n");
                    }
                    lastDir = 0;
                    sprintf(last_dir_buff, "%s", "None");
                    resetPasswordAttempt();
                    drawScreen = 1;
                }
                break;
            case MAINSCREEN:
                //See LOCKSCREEN for further clarifications
                if (drawScreen) {
                    tft_setCursor(20, 50);
                    tft_setTextColor(ILI9341_BLACK);
                    tft_writeString(buffer);

                    tft_setCursor(40, 100);
                    tft_setTextColor(main_option == 0 ? ILI9341_GREEN : ILI9341_WHITE);
                    tft_writeString("Lock Device\n");
                    tft_setTextColor(main_option == 1 ? ILI9341_GREEN : ILI9341_WHITE);
                    tft_writeString("Reset Password\n");

                    drawScreen = 0;
                }
                //Use either up or down to navigate the display
                if (drawScreen || actionCode == NORTH || actionCode == SOUTH) {
                    main_option = main_option ? 0 : 1;
                    tft_setCursor(40, 100);
                    tft_setTextColor(main_option == 0 ? ILI9341_GREEN : ILI9341_WHITE);
                    tft_writeString("Lock Device\n");
                    tft_setTextColor(main_option == 1 ? ILI9341_GREEN : ILI9341_WHITE);
                    tft_writeString("Reset Password\n");

                }
                //Single click for selection
                if (actionCode == SINGLE) {
                    tft_setTextColor(ILI9341_WHITE);
                    tft_fillScreen(ILI9341_BLACK);
                    tft_setCursor(0, 0);
                    c_screen = main_option == 0 ? LOCKSCREEN : NEWPASSSCREEN;
                    lastDir = 0;
                    sprintf(last_dir_buff, "%s", "None");
                    drawScreen = 1;
                    password_index = 0;
                }
                break;
            case NEWPASSSCREEN:
                //See LOCKSCREEN for further clarifications
                //Pattern selection heavily mimics functionality
                if (drawScreen) {
                    tft_writeString("Enter a new PSWRD\n");
                    resetPassword();

                    tft_setCursor(50, 120);
                    tft_writeString("Last Dir:");
                    tft_setCursor(100, 150);
                    tft_writeString(last_dir_buff);

                    drawScreen = 0;
                }
                
                if (actionCode > HOLD) {
                    if (password_index < MAXPASSLEN) {


                        tft_setCursor(100, 150);
                        tft_setTextColor(ILI9341_BLACK);
                        tft_writeString(last_dir_buff);

                        tft_setCursor(100, 150);
                        tft_setTextColor(ILI9341_WHITE);
                        sprintf(last_dir_buff, "%c", getDirChar(actionCode));
                        tft_writeString(last_dir_buff);

                        lastDir = actionCode;
                    }
                } else if (actionCode == DOUBLE && lastDir != 0) {
                    tft_setCursor(20, 50);
                    sprintf(pass_buff, "%s%s", pass_buff, last_dir_buff);
                    tft_writeString(pass_buff);
                    password[password_index++] = lastDir;

                    tft_setCursor(100, 150);
                    tft_setTextColor(ILI9341_BLACK);
                    tft_writeString(last_dir_buff);

                    tft_setCursor(100, 150);
                    tft_setTextColor(ILI9341_WHITE);
                    lastDir = 0;
                    sprintf(last_dir_buff, "%s", "None");
                    tft_writeString(last_dir_buff);
                } else if (actionCode == SINGLE && password_index > 0) {
                    tft_fillScreen(ILI9341_BLACK);
                    password_index = 0;
                    c_screen = LOCKSCREEN;
                    lastDir = 0;
                    sprintf(last_dir_buff, "%s", "None");
                    tft_setCursor(0, 0);
                    drawScreen = 1;
                }
                break;
        }
        actionCode = OFF;
    }

    PT_END(pt);
}

/* This thread will continuously poll the UART pin to determine whether or not
 * the user has sent another command/action.
 * Authors: Craig Lombardo & Rachel Szteinberg '17
 */
static PT_THREAD(protothread_read_UART(struct pt *pt)) {
    PT_BEGIN(pt);

    //Clear the reset on the chip
    PT_YIELD_TIME_msec(50);
    mPORTBClearBits(BIT_7);

    while (1) {

        PT_YIELD_TIME_msec(50);

        PT_SPAWN(pt, &pt_input, PT_GetSerialBuffer(&pt_input));
        //We do not directly write to the actionCode; however, you may do so if
        //you wish
        sscanf(PT_term_buffer, " %i ", &buffInput);

        actionCode = buffInput;

        //We never set this back to 0
        connection_established = 1;

    }

    PT_END(pt);
}

/* This main method sets up the device for pairing
 * Authors: Craig Lombardo & Rachel Szteinberg '17
 */
void main(void) {
    SYSTEMConfigPerformance(PBCLK);

    ANSELA = 0;
    ANSELB = 0;
    CM1CON = 0;
    CM2CON = 0;

    tft_setTextSize(1);
    tft_setTextColor(ILI9341_WHITE);

    configureADC();

    //RA3 = Red LED
    //RA4 = Blue LED
    //RB4 = Green LED
    mPORTASetPinsDigitalOut(BIT_3 | BIT_4);
    mPORTBSetPinsDigitalOut(BIT_4);

    mPORTAClearBits(BIT_3 | BIT_4);
    mPORTBClearBits(BIT_4);

    //RA2 = 
    mPORTASetPinsAnalogIn(BIT_2);

    //RB7 = BLE reset
    mPORTBSetPinsDigitalOut(BIT_7);

    PT_setup();

    INTEnableSystemMultiVectoredInt();

    PT_INIT(&pt_draw_screen);
    PT_INIT(&pt_input);
    PT_INIT(&pt_read_UART);

    tft_init_hw();
    tft_begin();
    tft_fillScreen(ILI9341_BLACK);
    tft_setRotation(3);
    tft_setTextSize(2);
    tft_setTextColor(ILI9341_WHITE);

    resetPassword();
    resetPasswordAttempt();

    //reset BLE chip
    mPORTBSetBits(BIT_7);

    sprintf(last_dir_buff, "%s", "None");

    while (1) {
        PT_SCHEDULE(protothread_draw_screen(&pt_draw_screen));
        PT_SCHEDULE(protothread_read_UART(&pt_read_UART));
    }
}